//--------------------------------------------------------------------------------------------------------------------
// Kitchen Alarm: Pololu A-STAR 328PB + FC16 Matrix LCD + RTC DS1307 + Speaker + Rotary knob
// Developed by:  Henk Beekhof
// License:       See license.txt
// -------------------------------------------------------------------------------------------------------------------
// LCD Module                 
// LC16 CLK  = 17 D13 (nano) == PB5 (A*) == SPI SCK                              
// LC16 DIN  = 15 D11 (nano) == PB3 (A*) == SPI MOSI
// LC16 CS   = 14 D10 (nano) == PB2 (A*) == SPI SS                                                  Vcc = 5V, GND = 0V
//--------------------------------------------------------------------------------------------------------------------
// Rotary enable 1 D1 (nano) : D7 = PD7 (A*) : requires internal pullup
// Rotary enable 2 D2 (nano) : D8 = PB0 (A*) : requires internal pullup
// Rotary button 4 D4 (nano) : D4 = PD4 (A*) : requires internal pullup  0=button pressed 
// -------------------------------------------------------------------------------------------------------------------
// Speaker sound 3 D3 (nano) : D3 = PD3 (A*) : PWM output
// Speaker on    5 D5 (nano) : D5 = PD5 (A*) : Amplifier on/off
//--------------------------------------------------------------------------------------------------------------------
// RTC DS1307 SQ    == D2 (nano) : PD2 (A*) : INT0  RTC square wave output, set to 1Hz, used as interrupt
// RTC DS1307 SCL   == A5 (nano) : PC5 (A*) : SCL
// RTC DS1307 SDA   == A4 (nano) : PC4 (A*) : SDA                                                  Vcc = 3V3, GND = 0V
//--------------------------------------------------------------------------------------------------------------------
// EEPROM Content Specification:
// ==========================================================================================
// Address | Data                   | Description
// ==========================================================================================
//      0  | TIMER_VALID_CODE       | If this code is available then a timer time is stored
//      1  | Timer h                | Timer time: hours byte
//      2  | Timer m                | Timer time: minutes byte
//      3  | Timer s                | Timer time: seconds byte
//      4  | Correction             | Time correction in seconds due to inaccurate RTC module
//      5  | LastCorrectTime YY msb | Time of the last correction: Year (MSB)
//      6  | LastCorrectTime YY lsb | Time of the last correction: Year (LSB)
//      7  | LastCorrectTime MM     | Time of the last correction: Month
//      8  | LastCorrectTime DD     | Time of the last correction: Day
//      9  | Timer sound            | Defines the timer sound
//     10  | ALARM_VALID_CODE       | If this code is available then an alarm time is stored
//     11  | Alarm 1 h              | Alarm time: hours byte
//     12  | Alarm 1 m              | Alarm time: minutes byte
//     13  | Alarm 1 s              | Alarm time: seconds byte
//     14  | Alarm 1 On/Off         | Alarm mode: 0x01=On, any other value is off
//     15  | Alarm 1 sound          | Defines the alarm sound
//     16  | Alarm 1 repeat         | Defines how many times the alarm is repeated
//     17  | Dark mode On/Off       | Dark mode: 0x01=On, any other value is off
//--------------------------------------------------------------------------------------------------------------------
// Notes related to library usage:
// 1) Eeprom24C32_64.h      This library does not work. Random multi byte reads do not honer the specified address
// 2) RTCLib                It does not turn on the square wave output when it gets disabled, only battery and power
//                          removal results in a hard reset in the chip, which turns on the clock. 
//                          This CH (Clock Hold) setting should be part of the RTCLib, it assumes the clock is running 
//                          To correct this it is checked in the Time.cpp file, see: Time_Interface::init()
//                          See also the MaximRTC chip DS1307 datasheet and search for 'Clock Hold'.
//--------------------------------------------------------------------------------------------------------------------
// Notes on the hardware
// The RTC module had and inaccuracy of rougly 12 seconds per day, this is not acceptable.
// The software includes a correction to 'fix' this issue, however the real problem in the HW boards is found.
// On the internet you find articles that the RTC chip is a 'copy' which does not have the correct internal capacitors
// I replaced the chip with RTC chips from Maxim, the result was the same. These chips require a certian load capacitance
// of the crystal. I bought serveral Xtals with different load cap specs. The 'wrong' Xtals (cap not matching the RTC chip)
// had indeed an offset in frequency. The Xtal with matching cap spec of 12.5 pF resulted in an accurate RTC board. 
// Thus replacing the RTC chip is not required, replacing the Xtal has huge impact on the accuracy, and fixes the issue.
//--------------------------------------------------------------------------------------------------------------------
#include <MD_MAX72xx.h>
#include <SPI.h>
#include <Wire.h>
#include <AT24CX.h>
#include "defs.h"
#include "Time.h"
#include "RTClib.h"
#include "Rotary_wemos.h"
#include "font3x5.h"
#include "font4x7_digits.h"
#include "pitches.h"
#include <avr/pgmspace.h>

//--------------------------------------------------------------------------------------------------------------------
// Specification of alarm sounds
// note durations: 4 = quarter note, 8 = eighth note, actual play time is 1000ms/<this number>
int alarm_1[] = {
  NOTE_C4, NOTE_C4, NOTE_C4, NOTE_C7, NOTE_C6, NOTE_C5, NOTE_C4, NOTE_C4, NOTE_C4
};
int alarm_1_durations[] = {
  4, 4, 4, 6, 6, 6, 4, 4, 4
};

int alarm_2[] = {
  NOTE_G4, NOTE_G4, NOTE_G4, NOTE_C5, NOTE_E4, NOTE_E4, NOTE_E4, NOTE_G4, NOTE_C4, NOTE_C4, NOTE_D4, NOTE_D4, NOTE_G4, NOTE_E4,
  NOTE_A4, NOTE_F4, NOTE_F4, NOTE_F4, NOTE_G4, NOTE_E4, NOTE_E4, NOTE_E4, NOTE_F4, NOTE_D4, NOTE_D4, NOTE_D4, NOTE_C4, NOTE_E4, NOTE_G4, NOTE_G4
};
int alarm_2_durations[] = {
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 2,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4
};

// notes in the melody:
int alarm_3[] = {
NOTE_G4, NOTE_G5, NOTE_FS5, NOTE_G5, NOTE_FS5, NOTE_D5, NOTE_E5, NOTE_C4, NOTE_C4,
NOTE_C4, NOTE_B4, NOTE_A4, NOTE_B4, NOTE_G4, NOTE_G5, NOTE_FS5, NOTE_G5, NOTE_A5, NOTE_FS5,
NOTE_E5, NOTE_D5, NOTE_E5, NOTE_D5, NOTE_B4, NOTE_C4, NOTE_C4, NOTE_C4, NOTE_B4, NOTE_A4,
NOTE_B4, NOTE_G4
};
 
int alarm_3_durations[] = {
4, 2, 16, 16, 16, 16, 2, 8, 8, 16, 16, 16, 16, 4, 2, 16, 16, 16, 16, 2, 16, 16, 16, 16,
8, 8, 16, 16, 16, 16, 2
};

int alarm_4[] = {
  NOTE_E4, NOTE_E4, REST, NOTE_E4, REST, NOTE_C4, NOTE_E4, REST, NOTE_G4, REST, REST, NOTE_G3, REST,
  NOTE_C4, REST, REST, NOTE_G3, REST, NOTE_E3, REST, REST, NOTE_A3, REST, NOTE_B3, REST, NOTE_AS3, NOTE_A3, REST,
  NOTE_G3, NOTE_E4, NOTE_G4, NOTE_A4, REST, NOTE_F4, NOTE_G4, REST, NOTE_E4, REST, NOTE_C4, NOTE_D4, NOTE_B3, REST
};
int alarm_4_durations[] = {
  4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 4, 2, 2,
  4, 4, 4, 4, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,  
  4, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2
};

//--------------------------------------------------------------------------------------------------------------------
// Instantiate the display
MD_MAX72XX mx = MD_MAX72XX(HARDWARE_TYPE, SPI_SS, MAX_DEVICES);

// Instantiate the RTC interface
Time_Interface RTCTime;   

// Instantiate the EEPROM interface
AT24C32 eeprom;

// Instantiate the rotary
Rotary R1(ROT_ENC1_PIN, ROT_ENC2_PIN, false);

//--------------------------------------------------------------------------------------------------------------------
// Definition of global variables
struct setTimeFormat {
  int h;                                  // Hours
  int m;                                  // Minutes
  int s;                                  // Seconds
};
setTimeFormat timerSetTime1;              // Time to count down for the kitchen timer
setTimeFormat alarmSetTime1;              // Time to compare to the actual time resulting in an alarm
volatile bool interrupt_1Hz_Flag;         // Set to true on each 1Hz interrupt to signal the 1Hz actions                             
unsigned char m0p;                        // Seconds are from 00 to 59, the lowest digit is represented in m0, m0p is the previous value
                                          // Usage: Set m0p outside the range 0..59 to force the hours and minutes to be updated on the display
unsigned char time_mode;                  // Specifies display format of time :== {TIME_MODE_LARGE, TIME_MODE_SMALL_HHMM, TIME_MODE_SMALL_HHMMSS}
char correction;                          // For inaccurate RTC modules a correction of N seconds per day is supported
bool correction_done, correction_active;  // Indicate if the time correction is done or active
char alarm1_on;                           // When set, the alarm is on
char dark_on;                             // When set, the dark mode is enabled (display is off, press the rotary to light up the display)
char dark_on_countdown;                   // Counts the duration the display is shown before it returns to dark mode
bool dark_on_and_show;                    // Indicates the state/period when the display is on in the dark mode

//--------------------------------------------------------------------------------------------------------------------
// Specify 'one liner' messages. Stored in Flash
const char string_0[]  PROGMEM = "Your text 1.";
const char string_1[]  PROGMEM = "Your text 2.";

//--------------------------------------------------------------------------------------------------------------------
// Specify sets of birthdays. One set per 'product==clock' targeting the clock owner. Stored in Flash
const char birthday_1A[] PROGMEM = {"John"};
const char birthday_2A[] PROGMEM = {"Ella"};
const char birthday_3A[] PROGMEM = {"Mia"};
//const char birthday_1B[] PROGMEM = {"Julia"};

const char *const textItems[] PROGMEM = {birthday_1A, birthday_2A, birthday_3A};
//const char *const textItems[] PROGMEM = {birthday_1B};


//--------------------------------------------------------------------------------------------------------------------
// Read the EEPROM and display data over the serial interface for debugging purposes
//--------------------------------------------------------------------------------------------------------------------
void read_display_eeprom_data() {
  byte eepromData[17];
  char buf[40];
  
  // Read the EEPROM
  eeprom.read(0, eepromData, 17);

  sprintf(buf, "%3d: TIMER_VALID_CODE", eepromData[0]);       Serial.println(buf);    // If this code is available then a timer time is stored
  sprintf(buf, "%3d: Timer h", eepromData[1]);                Serial.println(buf);    // Timer time: hours byte
  sprintf(buf, "%3d: Timer m", eepromData[2]);                Serial.println(buf);    // Timer time: minutes byte
  sprintf(buf, "%3d: Timer s", eepromData[3]);                Serial.println(buf);    // Timer time: seconds byte
  sprintf(buf, "%3d: Correction", eepromData[4]);             Serial.println(buf);    // Time correction in seconds due to inaccurate RTC module
  sprintf(buf, "%3d: Year MSB", eepromData[5]);               Serial.println(buf);    // Time of the last correction Year
  sprintf(buf, "%3d: Year LSB", eepromData[6]);               Serial.println(buf);    // Time of the last correction Year
  sprintf(buf, "%3d: Month", eepromData[7]);                  Serial.println(buf);    // Time of the last correction Month
  sprintf(buf, "%3d: Day", eepromData[8]);                    Serial.println(buf);    // Time of the last correction Day
  sprintf(buf, "%3d: Timer sound", eepromData[9]);            Serial.println(buf);    // Defines the timer sound
  sprintf(buf, "%3d: ALARM_VALID_CODE", eepromData[10]);      Serial.println(buf);    // if this code is available then an alarm time is stored
  sprintf(buf, "%3d: Alarm 1 h", eepromData[11]);             Serial.println(buf);    // Alarm time: hours byte
  sprintf(buf, "%3d: Alarm 1 m", eepromData[12]);             Serial.println(buf);    // Alarm time: minutes byte
  sprintf(buf, "%3d: Alarm 1 s", eepromData[13]);             Serial.println(buf);    // Alarm time: seconds byte
  sprintf(buf, "%3d: Alarm 1 On", eepromData[14]);            Serial.println(buf);    // 0x01=On, any other value is off
  sprintf(buf, "%3d: Alarm 1 sound", eepromData[15]);         Serial.println(buf);    // Defines the alarm sound
  sprintf(buf, "%3d: Alarm 1 repeat", eepromData[16]);        Serial.println(buf);    // Defines how many times the alarm is repeated
}
  
//--------------------------------------------------------------------------------------------------------------------
// Text messages
//--------------------------------------------------------------------------------------------------------------------
void show_oneliner() {
  char buf[75];
  int idx;
  
  idx = (int)random(4);          // random(N): Generates a number between 0 and N-1
  strcpy_P(buf, (char *)pgm_read_word(&(textItems[idx])));
  scroll_string_by_shift_left(buf, 30);
}

//--------------------------------------------------------------------------------------------------------------------
// Interrupt Service Routine: Set the global 1Hz flag
//--------------------------------------------------------------------------------------------------------------------
void Int_1Hz_Routine() {
  interrupt_1Hz_Flag = true;
}

//--------------------------------------------------------------------------------------------------------------------
// Print a string from the default font on location x, return value is first free column after  
// the string is printed, including an empty column for spacing
//--------------------------------------------------------------------------------------------------------------------
int print_string(char *pMsg, unsigned char x) {
  uint8_t cBuf[8];
  uint8_t charWidth, idx;
  
  while (*pMsg != '\0') {
    // Retrieve the next character form the font file
    charWidth = mx.getChar(*pMsg++, sizeof(cBuf)/sizeof(cBuf[0]), cBuf);
    for (idx = 0; idx < charWidth; idx++) {
      mx.setColumn(x--, cBuf[idx]);
    }
    // Add empty column
    mx.setColumn(x--, 0);
  }
  return x;
}

//--------------------------------------------------------------------------------------------------------------------
// Scroll a string from the default font from location 0 to 31, from right to left  
//--------------------------------------------------------------------------------------------------------------------
void scroll_string_by_shift_left(char *pMsg, int xSpeed) {
  uint8_t cBuf[8];
  uint8_t charWidth, idx;

  mx.clear();
  while (*pMsg != '\0') {
    // Retrieve the next character form the font file
    charWidth = mx.getChar(*pMsg++, sizeof(cBuf)/sizeof(cBuf[0]), cBuf);
    for (idx = 0; idx < charWidth; idx++) {
      mx.setColumn(0, cBuf[idx]);
      mx.transform(MD_MAX72XX::TSL);
      delay(xSpeed);
    }
    mx.setColumn(0, 0);
    mx.transform(MD_MAX72XX::TSL);
    delay(xSpeed);
  }
  
  // Shift all out
  for (idx = 0; idx < 31; idx++) {
    mx.setColumn(0, 0);
    mx.transform(MD_MAX72XX::TSL);
    delay(xSpeed);
  }
  
  mx.clear();
  m0p = 99;
}

//--------------------------------------------------------------------------------------------------------------------
// Print empty columns to clear the display from column x by the specified lenght 
// -------------------------------------------------------------------------------------------------------------------
void print_clear(unsigned char x, unsigned char len) {
  // Printing direction is from left to right, thus decrement the position starting at column x
  char xStop;

  xStop = 0;
  if ((x - len + 1) >= 0) {
    xStop = x - len + 1;
  }
  for (char pos = x; pos >= xStop; pos--) {
    mx.setColumn(pos, 0);
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print 3x5 char on {col = x (0=right), row = y (0=top)}, overwrites the full column 
// -------------------------------------------------------------------------------------------------------------------
void print_string3(char *pMsg, unsigned char x, unsigned char y) {
  unsigned char charIdx, charColData;

  while (*pMsg != '\0') {
    // Calculate the font array offset for this character
    // The font does not contain all chars from the ascii table, space ascii = 32, at location 0 => (value-32)*3
    // Chars out of this range are printed as a space character (charIdx = 0)
    charIdx = 0;
    if ((*pMsg >= 32) && (*pMsg <=126)) {
      charIdx = (*pMsg-32) * 3;
    }
    
    for (byte col = 0; col <= 2; col++) {
      charColData = (pgm_read_byte_near(&Font3x5[charIdx++]) << y);
      mx.setColumn(x--, charColData);
    }
    x--;
    pMsg++;
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print 3x5 digit on {col = x (0=right), row = y (0=top)}, overwrites the full column 
// -------------------------------------------------------------------------------------------------------------------
void print_digit3(unsigned char digit, unsigned char x, unsigned char y) {
  unsigned char charColData;
  
  // Calculate the font array offset for this character
  unsigned char charIdx = (digit + 16) * 3;
  for (byte col = 0; col <= 2; col++) {
    charColData = (pgm_read_byte_near(&Font3x5[charIdx++]) << y);
    mx.setColumn(x--, charColData);
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print 4x7 digit on {col = x (0=right), row = y (0=top)}, overwrites the full column 
// -------------------------------------------------------------------------------------------------------------------
void print_digit4(unsigned char digit, unsigned char x, unsigned char y) {
  unsigned char charColData;

  // Calculate the font array offset for this character
  unsigned char charIdx = digit << 2;
  for (byte col = 0; col <= 3; col++) {
    charColData = (pgm_read_byte_near(&Font4x7[charIdx++]) << y);
    mx.setColumn(x--, charColData);
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print 3x5 digit on {col = x (0=right), row = y (0=top)}, writes dot positions 
// -------------------------------------------------------------------------------------------------------------------
void print_digit3dots(unsigned char digit, unsigned char x, unsigned char y) {
  unsigned char charColData;

  // Calculate the font array offset for this character
  unsigned char charIdx = (digit + 16) * 3;
  for (byte col = 0; col <= 2; col++) {
    charColData = pgm_read_byte_near(&Font3x5[charIdx++]);
    for (char row = 0; row <= 4; row++) {
      if (charColData & 0x01) {
        mx.setPoint(y + row, x - col, 1);
      } else {
        mx.setPoint(y + row, x - col, 0);
      }
      charColData = charColData >> 1;  
    }
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print 4x7 digit on {col = x (0=right), row = y (0=top)}, writes dot positions 
// -------------------------------------------------------------------------------------------------------------------
void print_digit4dots(unsigned char digit, unsigned char x, unsigned char y) {
  unsigned char charColData;

  // Calculate the font array offset for this character
  unsigned char charIdx = digit << 2;
  for (char col = 0; col <= 3; col++) {
    charColData = pgm_read_byte_near(&Font4x7[charIdx++]) >> 1;
    for (char row = 0; row <= 6; row++) {
      if (charColData & 0x01) {
        mx.setPoint(y + row, x - col, 1);
      } else {
        mx.setPoint(y + row, x - col, 0);
      }
      charColData = charColData >> 1;  
    }
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Show the alarm indicator (alarm 1 on/off visualized by the top-right x,y={0,0} led on/off)
// -------------------------------------------------------------------------------------------------------------------
void show_alarm_onoff() {
  if (alarm1_on) {
    mx.setPoint(0, 0, 1);  
  } else {
    mx.setPoint(0, 0, 0);  
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print the birthday message, input string is a name 
// -------------------------------------------------------------------------------------------------------------------
void show_birthday_message(char *buf_name) {
  char buf[64];
  strcpy(buf, buf_name);
  strcat(buf, ", gefeliciteerd met je verjaardag !");   // English change this to ", happy birthday !"
  scroll_string_by_shift_left(buf, 30);
  scroll_string_by_shift_left(buf, 30);
}

//--------------------------------------------------------------------------------------------------------------------
// Play the alarm sound sequence 
// -------------------------------------------------------------------------------------------------------------------
bool play_alarm(byte alarmSound) {
  int idx, noteDuration, pauseDuration, nrOfItems;
  bool stopAlarm = false;
  
  digitalWrite(SOUND_ON, 1);
  
  switch (alarmSound) {
    //-------------------------------------------------------------------------------------------
    case 0: 
      for (idx=80; idx<500; idx+=4) {
        tone(SOUND_OUTPUT, idx, 20);
        delay(15);
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
      for (idx=500; idx>100; idx-=12) {
        tone(SOUND_OUTPUT, idx, 15);
        delay(10);
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
      break;
      
    //-------------------------------------------------------------------------------------------  
    case 1:
      float sinVal;
      int toneVal;
      for (int x=0; x<140; x++) {
        // convert degrees to radians then obtain sin value
        sinVal = (sin(x*(3.1412/180)));
        toneVal = 2000+(int(sinVal*1000));
        tone(SOUND_OUTPUT, toneVal);
        delay(4);
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
      break;
      
    //-------------------------------------------------------------------------------------------   
    case 2:
      for (idx = 0; idx < 30; idx++) {      
        // Note duration is 1 second divided by duration
        int noteDuration = 1000 / alarm_2_durations[idx];   
        tone(SOUND_OUTPUT, alarm_2[idx], noteDuration);
        
        pauseDuration = noteDuration * 1.20;     
        delay(pauseDuration);
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
      break;

    //-------------------------------------------------------------------------------------------   
    case 3:
      for (idx = 0; idx < 3; idx++) {      
        // Note duration is 1 second divided by duration
        int noteDuration = 1000 / alarm_2_durations[idx];   
        tone(SOUND_OUTPUT, NOTE_E5, 500);
        
        delay(700);
        
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
      break;
      
    //-------------------------------------------------------------------------------------------   
    case 4:
      nrOfItems = sizeof(alarm_3) / sizeof(alarm_3[0]);
      for (idx = 0; idx < nrOfItems; idx++) {      
        // Note duration is 1 second divided by duration
        int noteDuration = 2000 / alarm_3_durations[idx];   
        tone(SOUND_OUTPUT, alarm_3[idx], noteDuration);
        
        pauseDuration = noteDuration * 1.30;     
        delay(pauseDuration);
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
      break;  
      
    //-------------------------------------------------------------------------------------------
    default:
      for (idx = 0; idx < 9; idx++) {      
        // Note duration is 1 second divided by duration
        int noteDuration = 500 / alarm_1_durations[idx];   
        tone(SOUND_OUTPUT, alarm_1[idx], noteDuration);
        
        pauseDuration = noteDuration * 1.50;     
        delay(pauseDuration);
        if (!digitalRead(ROT_BUTTON_PIN)) {
          stopAlarm = true;
          break;
        }
      }
  }

  // Stop sound      
  noTone(SOUND_OUTPUT);  
  digitalWrite(SOUND_ON, 0);
  
  while (!digitalRead(ROT_BUTTON_PIN));  // If the button was pressed then WAIT until release
  
  return stopAlarm;
}

//--------------------------------------------------------------------------------------------------------------------
// Print time HH:MM:ss 
// -------------------------------------------------------------------------------------------------------------------
void print_time(unsigned char displayMode) {
  int h, m, s, yr, mo, da;
  unsigned char h1, h0, m1, m0, s1, s0, alarmCount, alarmSound;
  char buf[75];
  bool stopAlarm;
  
  // Fetch the time
  RTCTime.readHMS(&h, &m, &s);

  // Check the alarm time
  if (alarm1_on) {
    if ((h == alarmSetTime1.h) && (m == alarmSetTime1.m) && (s == alarmSetTime1.s)) {
      alarmSound = eeprom.read(ALARM1_SOUND_MODE_ADDRESS);
      alarmCount = eeprom.read(ALARM1_REPEAT_ADDRESS);
      for (int i=0; i<alarmCount; i++) {
        mx.clear();
        print_string("Alarm", 28);
        stopAlarm = play_alarm(alarmSound);
        if (stopAlarm) break;
        delay(250);
        mx.clear();
        delay(250);
      }
      mx.clear();
      m0p = 99;
      return;
    }
  }

  // If the dark mode is on, the alarm is checked already, so the rest of this routine to display can be skipped
  if ((dark_on == 1) && (dark_on_and_show == false)) return;

  // Check if we are on an exact new hour HH:00:00  Shows message on whole hour
  if ((m == 0) && (s == 0)) {
    
    // Check if we are at a certain date to display a date related message
    RTCTime.readYMD(&yr, &mo, &da);

    // Birthday of: John (Jan 27), Ella (Dec 8), Mia (Jun 9)
    if ((mo == 1) && (da == 27)) {
      strcpy_P(buf, (char *)pgm_read_word(&(textItems[0])));
      show_birthday_message(buf);
    } else if ((mo == 12) && (da == 8)) {
      strcpy_P(buf, (char *)pgm_read_word(&(textItems[1])));
      show_birthday_message(buf);
    } else if ((mo == 6) && (da == 9)) {
      strcpy_P(buf, (char *)pgm_read_word(&(textItems[2])));
      show_birthday_message(buf);

    // Birthday of: Julia (Oct 28)
//    if ((mo == 10) && (da == 28)) {
//      strcpy_P(buf, (char *)pgm_read_word(&(textItems[0])));
//      show_birthday_message(buf);

    } else {
      // Select here what you want to display on every hour crossing, a one liner text, or something else
      //show_oneliner();
      sprintf(buf, "Het is %d uur.", h);         // Dutch
      //sprintf(buf, "It's %d o'clock", h);      // English
      scroll_string_by_shift_left(buf, 30);
      scroll_string_by_shift_left(buf, 30);
    }
    mx.clear();
    m0p = 99;
    return;
  }

  // Correct time due to inaccuracy of RTC module, done once per day
  if (h==23 && m==59 && s==59) {
    
    // Check if the correction is already done, if not, start it
    if (!correction_done) {

      // Make sure the correction data is loaded once
      if (!correction_active) {
        correction = eeprom.read(CORRECTION_ADDRESS);
        correction_active = true;
      }
      
      // When correction is not zero, the count down and save that value in the RTC
      if (correction > 0) {
        s--;
        correction--;
        RTCTime.setHMS(h, m, s);
      }

      // When the correction equals zero, we are done, set correction_done to skip correction 
      // since the time is still the same due to keeping the clock at this time (s--)
      if (correction == 0) {
        correction_active = false;
        
        // Only if time is corrected activate the done activities
        if (s == 58) correction_done = true;
      }
      
    } else {
      // At the end of the correction this state is entered to turn off the correction algorithm
      correction_done = false;
    }
  }

  // Extract the hour, minute and second digit values  
  s0 = s % 10;
  s1 = s / 10;
  m0 = m % 10;
  m1 = m / 10;
  h0 = h % 10;
  h1 = h / 10;
  
  if (!correction_active) {
    if (displayMode == TIME_MODE_LARGE) {
      // Display the time
      print_digit3(s0, SEC_LSB, 2);
      print_digit3(s1, SEC_MSB, 2);
    
      // On a change of the minutes, update the hour, minutes and separation colons
      if (m0 != m0p) {
        // Separation colons
        mx.setColumn(SEP_MIN_SEC, 0x28);
        mx.setColumn(SEP_HRS_MIN, 0x44);
      
        print_digit4(m0, MIN_LSB, 1);
        print_digit4(m1, MIN_MSB, 1);
        print_digit4(h0, HRS_LSB, 1);
        print_digit4(h1, HRS_MSB, 1);

        show_alarm_onoff();
      }
    }
    
    if (displayMode == TIME_MODE_SMALL_HHMMSS) {
      // Display the time
      print_digit3(s0, SEC_LSB_SMALL_HHMMSS, 2);
      print_digit3(s1, SEC_MSB_SMALL_HHMMSS, 2);

      // On a change of the minutes, update the hour, minutes and separation colons
      if (m0 != m0p) {
        // Separation colons
        mx.setColumn(SEP_MIN_SEC_SMALL_HHMMSS, 0x28);
        mx.setColumn(SEP_HRS_MIN_SMALL_HHMMSS, 0x28);
      
        print_digit3(m0, MIN_LSB_SMALL_HHMMSS, 2);
        print_digit3(m1, MIN_MSB_SMALL_HHMMSS, 2);
        print_digit3(h0, HRS_LSB_SMALL_HHMMSS, 2);
        print_digit3(h1, HRS_MSB_SMALL_HHMMSS, 2);

        show_alarm_onoff();
      }
    }
    
    m0p = m0;
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Fetch user input by rotary change
// displayText:    Info to display
// xStartPos:      Position where the info starts, range from 0 (rightmost) to 31 (leftmost)
// currentValue:   Pointer to the value to change
// min, max:       Range in which the current value can be changed, min <= current value <= max 
// -------------------------------------------------------------------------------------------------------------------
bool fetch_rotary_value(char *displayText, unsigned char xStartPos, int *currentValue, int min, int max) {
  bool initDone, valueChanged;
  int R1Value, R1ValueOld, xPos, dummy;
  char buf[8];
  
  R1.Reset();
  
  mx.clear();
  xPos = print_string(displayText, xStartPos);
  if (*currentValue < 100) {
    sprintf(buf, "%02d ", *currentValue);
  } else {
    sprintf(buf, "%04d ", *currentValue);
  }
  dummy = print_string(buf, xPos);

  valueChanged = false;
  initDone = false;
  while (digitalRead(ROT_BUTTON_PIN)) { 
    R1Value = R1.Read();
    if (R1.Changed()) {
      if (!initDone) {
        R1ValueOld = R1Value;
        initDone = true;
      }
      if (abs(R1Value - R1ValueOld) > 4) {
        //Serial.print("old:");
        //Serial.print(R1ValueOld);
        //Serial.print(" new:");
        //Serial.println(R1Value);
        if (R1Value > R1ValueOld) {
          if (*currentValue < max) {
            (*currentValue)++;
          } else {
            *currentValue = min;
          }
          valueChanged = true;
        } else {
          if (*currentValue > min) {
            (*currentValue)--;
          } else {
            *currentValue = max;
          }
          valueChanged = true;
        }
        R1ValueOld = R1Value; 

        sprintf(buf, "%02d", *currentValue);
        dummy = print_string(buf, xPos);
        print_clear(dummy, 10);
      }
    } 
  }
  while (!digitalRead(ROT_BUTTON_PIN));  // WAIT until release
  return valueChanged;
}

//--------------------------------------------------------------------------------------------------------------------
// Set the last known correction time to the current time in the EEPROM 
// -------------------------------------------------------------------------------------------------------------------
void set_last_correction_time() {
  DateTime currentTime;
  byte timeData[4];

  // Read the currrent time
  currentTime = RTCTime.readCurrentTime();

  timeData[0] = (byte)((currentTime.year() & 0xFF00) >> 8);
  timeData[1] = (byte)(currentTime.year() & 0x00FF);
  timeData[2] = (byte)currentTime.month();
  timeData[3] = (byte)currentTime.day();
  eeprom.write(CORRECTION_TIME_ADDRESS, timeData, 4);
}

//--------------------------------------------------------------------------------------------------------------------
// Get the last known correction time from EEPROM 
// -------------------------------------------------------------------------------------------------------------------
DateTime get_last_correction_time() {
  byte timeData[4];
  int year;
  DateTime t;
  
  // Read the EEPROM
  eeprom.read(CORRECTION_TIME_ADDRESS, timeData, 4);

  // Convert 
  year = (timeData[0] << 8) + timeData[1];
  t = DateTime(year, timeData[2], timeData[3]);

  return t;
}

//--------------------------------------------------------------------------------------------------------------------
// Set date YYYY:MM:DD 
// -------------------------------------------------------------------------------------------------------------------
void set_date() {
  int y, m, d, R1Value, R1ValueOld, xPos, dummy;
  unsigned char y3, y2, y1, y0, m1, m0, d1, d0;
  char buf[8];
  bool updateRequired;
  
  // Fetch the date
  RTCTime.readYMD(&y, &m, &d);
  R1.Reset();
  R1ValueOld = 0;
  
  // Get user input
  updateRequired = false;
  updateRequired |= fetch_rotary_value("Y:", 31, &y, 2000, 3000);
  updateRequired |= fetch_rotary_value("M:", 31, &m, 1, 12);
  updateRequired |= fetch_rotary_value("D:", 31, &d, 1, 31);

  // Set the date
  mx.clear();
  if (updateRequired) {
    RTCTime.setYMD(y, m, d);

    // Set the last known correction time to this new date
    set_last_correction_time();
  }

  // Update the display by forced update of all digits
  m0p = 99;
  print_time(time_mode);
}

//--------------------------------------------------------------------------------------------------------------------
// Set time HH:MM:ss 
// -------------------------------------------------------------------------------------------------------------------
void set_time() {
  int h, m, s, R1Value, R1ValueOld, xPos, dummy;
  unsigned char h1, h0, m1, m0, s1, s0;
  byte timerTime[4];                    //One byte for the valid code, the hours, the minutes and the seconds
  char buf[8];
  bool updateRequired;

  // Fetch the time
  RTCTime.readHMS(&h, &m, &s);
  R1.Reset();
  R1ValueOld = 0;
  
  // Get user input
  updateRequired = false;
  updateRequired |= fetch_rotary_value("HH:", 31, &h, 0, 23);
  updateRequired |= fetch_rotary_value("MM:", 31, &m, 0, 59);
  updateRequired |= fetch_rotary_value("SS:", 31, &s, 0, 59);

  // Set the time
  mx.clear();
  if (updateRequired) {
    RTCTime.setHMS(h, m, s);
  }

  // Update the display by forced update of all digits
  m0p = 99;
  print_time(time_mode);
}  

//--------------------------------------------------------------------------------------------------------------------
// Set alarm time HH:MM:ss 
// -------------------------------------------------------------------------------------------------------------------
void set_alarm_time() {
  byte alarmTime[4];                    //One byte for the hours, the minutes and the seconds
  int R1Value, R1ValueOld;
  bool updateRequired;

  // Read the alarm time and check if it's valid (valid = first byte equals 'ALARM_VALID_CODE')
  eeprom.read(ALARM1_ADDRESS, alarmTime, 4);
  
  //char buf[60];
  //sprintf(buf, "Read code:%d\nTime:%02d:%02d:%02d", alarmTime[0], alarmTime[1], alarmTime[2], alarmTime[3]);
  //Serial.println(buf);
  
  if (alarmTime[0] == ALARM_VALID_CODE) {
    alarmSetTime1.h = alarmTime[1];
    alarmSetTime1.m = alarmTime[2];
    alarmSetTime1.s = alarmTime[3];
  }

  // Get user input
  updateRequired = false;
  updateRequired |= fetch_rotary_value("HH:", 31, &alarmSetTime1.h, 0, 23);
  updateRequired |= fetch_rotary_value("MM:", 31, &alarmSetTime1.m, 0, 59);
  updateRequired |= fetch_rotary_value("SS:", 31, &alarmSetTime1.s, 0, 59);

  // If the alarm time is changed, then save it to the EEPROM
  if (updateRequired) {
    alarmTime[0] = ALARM_VALID_CODE;
    alarmTime[1] = alarmSetTime1.h;
    alarmTime[2] = alarmSetTime1.m;
    alarmTime[3] = alarmSetTime1.s;
    eeprom.write(ALARM1_ADDRESS, alarmTime, 4);
  }
} 

//--------------------------------------------------------------------------------------------------------------------
// Set the amount of times the alarm sound is repeated [1..250]
// -------------------------------------------------------------------------------------------------------------------
void set_alarm_count() {
  unsigned char count;
  int updateValue;
  bool updateRequired;
  
  // Read the alarm count
  count = eeprom.read(ALARM1_REPEAT_ADDRESS);
  if (count > 250) count = 5;   // Also the setup guarantees that the default of 0xFF is overwritten with value 5
  
  // Get user input
  updateValue = (int)count;
  updateRequired = fetch_rotary_value("#:", 31, &updateValue, 1, 250);
  
  // If the count is changed, then save it to the EEPROM
  if (updateRequired) {
    count = (unsigned char)updateValue;
    eeprom.write(ALARM1_REPEAT_ADDRESS, count);
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Set the LCD intensity [1..15]
// -------------------------------------------------------------------------------------------------------------------
void set_lcd_intensity() {
  int lcd_intensity;

  lcd_intensity = 1;
  fetch_rotary_value("LCD:", 31, &lcd_intensity, 1, 15);
  mx.control(0, mx.INTENSITY, lcd_intensity);
  mx.control(1, mx.INTENSITY, lcd_intensity);
  mx.control(2, mx.INTENSITY, lcd_intensity);
  mx.control(3, mx.INTENSITY, lcd_intensity);
}

//--------------------------------------------------------------------------------------------------------------------
// Turn the dark display mode on or off
// -------------------------------------------------------------------------------------------------------------------
void set_dark_onoff(boolean onoff) {
  dark_on = onoff;
  eeprom.write(DARK_ON_ADDRESS, onoff);

  // Turn off display
  if (dark_on == 1) {
    mx.clear();
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Print timer: hh:mm:ss 
// -------------------------------------------------------------------------------------------------------------------
void print_timer() {
  int h,m,s;
  unsigned char h1, h0, m1, m0, s1, s0;

  // The time is known in the variable: timerSetTime1.h : timerSetTime1.m : timerSetTime1.s
  RTCTime.readHMS(&h, &m, &s);

  // Extract the hour, minute and second digit values  
  s0 = timerSetTime1.s % 10;
  s1 = timerSetTime1.s / 10;
  m0 = timerSetTime1.m % 10;
  m1 = timerSetTime1.m / 10;
  h0 = timerSetTime1.h % 10;
  h1 = timerSetTime1.h / 10;

  // Display the timer
  print_digit3(s0, TIMER_SEC_LSB, 2);
  print_digit3(s1, TIMER_SEC_MSB, 2);

  // On a change of the minutes, update the hour, minutes and separation colons
  if (m0 != m0p) {
    // Separation colons
    mx.setColumn(TIMER_SEP_MIN_SEC, 0x28);
    mx.setColumn(TIMER_SEP_HRS_MIN, 0x28);
  
    print_digit3(m0, TIMER_MIN_LSB, 2);
    print_digit3(m1, TIMER_MIN_MSB, 2);
    print_digit3(h0, TIMER_HRS_LSB, 2);
    print_digit3(h1, TIMER_HRS_MSB, 2);
  }
  m0p = m0;
}

//--------------------------------------------------------------------------------------------------------------------
// Increment the timer 
// -------------------------------------------------------------------------------------------------------------------
bool timer_increment() {
  if (timerSetTime1.s < 59) {
    timerSetTime1.s++;
    return false;
  }
  if (timerSetTime1.m < 59) {
    timerSetTime1.s = 0;
    timerSetTime1.m++;
    return false;
  }
  if (timerSetTime1.h < 23) {
    timerSetTime1.s = 0;
    timerSetTime1.m = 0;
    timerSetTime1.h++;
    return false;
  }
  // Time is 00:00:00
  return true;
}

//--------------------------------------------------------------------------------------------------------------------
// Change the timer with a delta value in seconds, signed value to add or subtract
// -------------------------------------------------------------------------------------------------------------------
void timer_update_delta(int delta) {
  // Convert HH:MM:SS into a single number in seconds, use delta to create new time
  int t = timerSetTime1.h * 3600 + timerSetTime1.m * 60 + timerSetTime1.s + delta;
  if (t < 0) t=0;
  
  // Convert the new time into HH:MM:SS
  timerSetTime1.h = (int)(t / 3600);
  t -= timerSetTime1.h * 3600;
  timerSetTime1.m = (int)(t / 60);
  timerSetTime1.s = t - timerSetTime1.m * 60;
}

//--------------------------------------------------------------------------------------------------------------------
// Decrement the timer 
// -------------------------------------------------------------------------------------------------------------------
bool timer_decrement() {
  if (timerSetTime1.s > 0) {
    timerSetTime1.s--;
    // Check if 00:00:00 is reached
    if ((timerSetTime1.h == 0) && (timerSetTime1.m == 0) && (timerSetTime1.s == 0)) {
      return true;
    } else {
      return false;
    }
  }
  if (timerSetTime1.m > 0) {
    timerSetTime1.s = 59;
    timerSetTime1.m--;
    return false;
  }
  if (timerSetTime1.h > 0) {
    timerSetTime1.s = 59;
    timerSetTime1.m = 59;
    timerSetTime1.h--;
    return false;
  }
  // Time is 00:00:00
  return true;
}
      
//--------------------------------------------------------------------------------------------------------------------
// Timer function 
// -------------------------------------------------------------------------------------------------------------------
void process_timer() {
  byte timerTime[4];                    //One byte per: the valid code, the hours, the minutes and the seconds
  int R1Value, R1ValueOld;
  bool timerState, initDone, valueChanged, stopAlarm;
  byte alarmSound;                      //Timer sound
  unsigned int shift;
  unsigned long current_timestamp, previous_timestamp;
  unsigned long diff;
    
  // Read the timer time and check if it's valid (valid = first byte equals 'TIMER_VALID_CODE')
  eeprom.read(TIMER_ADDRESS, timerTime, 4);
  
  //char buf[60];
  //sprintf(buf, "Read code:%d\nTime:%02d:%02d:%02d", timerTime[0], timerTime[1], timerTime[2], timerTime[3]);
  //Serial.println(buf);
  
  if (timerTime[0] == TIMER_VALID_CODE) {
    timerSetTime1.h = timerTime[1];
    timerSetTime1.m = timerTime[2];
    timerSetTime1.s = timerTime[3];
  }

  // Display the current known timer time: timerSetTime 
  mx.clear();
  m0p = 99;
  print_timer();
  
  // Set a new time (rotate) --OR-- Start the timer (button)
  valueChanged = false;
  R1ValueOld = 0;
  R1.Reset();
  previous_timestamp = millis();
  
  // Button is not yet pressed, enter the while loop: 0=button is pressed
  while (digitalRead(ROT_BUTTON_PIN)) { 
    R1Value = R1.Read();
    if (R1.Changed()) {
      
      if (abs(R1Value - R1ValueOld) > 4) {
        // Get rotation speed info to use an exponential increase on faster rotation
        current_timestamp = millis();
        diff = current_timestamp - previous_timestamp;
          
        shift = 1;
        if (diff < 100) shift = 5;
        if (diff <  60) shift = 30;
        if (diff <  30) shift = 300;
        if (diff <  20) shift = 900;
        if (diff <  10) shift = 1800;
          
        if (R1Value > R1ValueOld) {
          timer_update_delta(shift);
        } else {
          timer_update_delta(-shift);
        }
        valueChanged = true;
        R1ValueOld = R1Value; 
        previous_timestamp = current_timestamp;

        // Update the changed value on the LCD
        print_timer();
      }
    }
  }
  
  // If the alarm time is changed, then save it to the EEPROM
  if (valueChanged) {
    timerTime[0] = TIMER_VALID_CODE;
    timerTime[1] = timerSetTime1.h;
    timerTime[2] = timerSetTime1.m;
    timerTime[3] = timerSetTime1.s;
    eeprom.write(TIMER_ADDRESS, timerTime, 4);
  }  

  // Check if the button is released
  while (!digitalRead(ROT_BUTTON_PIN));  //WAIT until button release

  // To prevent jumps, wait one 1Hz pulse
  while (!interrupt_1Hz_Flag) { /* WAIT unit we get in sync */ }
  interrupt_1Hz_Flag = false;

  // Execute the timer functionality: Count down
  timerState = false;
  while (!timerState) {
    // When the 1Hz flag is set, update the time and/or do other 1Hz related tasks
    if (interrupt_1Hz_Flag) {
      timerState = timer_decrement();
      print_timer();
      interrupt_1Hz_Flag = false;
    } 
  }

  alarmSound = eeprom.read(TIMER_SOUND_MODE_ADDRESS);
  for (int i=0; i<5; i++) {
    mx.clear();
    print_string("Ready", 30);
    stopAlarm = play_alarm(alarmSound);
    if (stopAlarm) break;
    delay(250);
    mx.clear();
    delay(250);
  }
  mx.clear();
  m0p = 99;
}

//--------------------------------------------------------------------------------------------------------------------
// Execute the menu to set the alarm mode
// -------------------------------------------------------------------------------------------------------------------
void set_sound(char selectedFunction) {
  char menuItems[5][4] = {"S1", "S2", "S3", "S4", "S5"};
  unsigned char idx = 0;
  unsigned char idxMax = 4;
  bool valueChanged;
  int R1Value, R1ValueOld;

  mx.clear();
  mx.setColumn(31, 0x3E);
  mx.setColumn(30, 0x1C);
  mx.setColumn(29, 0x08);
  print_string(menuItems[idx], 27);
  play_alarm(0);
  
  R1.Reset();
  R1ValueOld = 0;
  valueChanged = false;

  // As long as the button is not pressed, walk through the menu items, a button press selects the item
  while (digitalRead(ROT_BUTTON_PIN)) { 
    R1Value = R1.Read();
    if (R1.Changed()) {
      if (abs(R1Value - R1ValueOld) > 4) {
        if (R1Value > R1ValueOld) {
          if (idx < idxMax) idx++;
          valueChanged = true;
        } else {
          if (idx > 0) idx--;
          valueChanged = true;
        }
        R1ValueOld = R1Value; 

        mx.clear();
        mx.setColumn(31, 0x3E);
        mx.setColumn(30, 0x1C);
        mx.setColumn(29, 0x08);
        print_string(menuItems[idx], 27);

        play_alarm(idx);
      }
    } 
  }
  while (!digitalRead(ROT_BUTTON_PIN));  // WAIT until release
  
  mx.clear();

  if (selectedFunction == TIMER)  {
    eeprom.write(TIMER_SOUND_MODE_ADDRESS, idx);
  }
  if (selectedFunction == ALARM1) {
    eeprom.write(ALARM1_SOUND_MODE_ADDRESS, idx);
  }
}

//--------------------------------------------------------------------------------------------------------------------
// Turn the alarm setting on (activate alarm) or off (deactivate alarm)
// -------------------------------------------------------------------------------------------------------------------
void set_alarm_onoff(boolean onoff) {
  alarm1_on = onoff;
  eeprom.write(ALARM1_ON_ADDRESS, onoff);
}

//--------------------------------------------------------------------------------------------------------------------
// Display the stopwatch time
// -------------------------------------------------------------------------------------------------------------------
void displayStopwatchTime(unsigned long dTime) {
  unsigned int mins, secs, mils;
  unsigned char min1, min0, sec1, sec0, mil2, mil1, mil0;

  // Separation colons, Display format is 00:00:000 [MM:SS:mmm]
  mx.setColumn(SW_SEP0, 0x28);
  mx.setColumn(SW_SEP1, 0x28);
  
  // Extract the minute, second and millisecond values 
  secs = dTime / 1000;
  mins = secs / 60; 
  mils = dTime % 1000;

  // Extract the individual numbers for minutes, seconds and milliseconds
  min1 = mins / 10;
  min0 = mins % 10;
  sec1 = (secs % 60) / 10;
  sec0 = (secs % 60) % 10;
  mil2 = (mils / 100);
  mils = mils - mil2*100;
  mil1 = mils / 10;
  mil0 = mils % 10;
  
  // Display stopwatch time
  print_digit3(mil0, MSEC0, 2);
  print_digit3(mil1, MSEC1, 2);
  print_digit3(mil2, MSEC2, 2);
  print_digit3(sec0, SEC0, 2);
  print_digit3(sec1, SEC1, 2);
  print_digit3(min0, MIN0, 2);
  print_digit3(min1, MIN1, 2);
}

//--------------------------------------------------------------------------------------------------------------------
// Stopwatch function
// -------------------------------------------------------------------------------------------------------------------
void stopwatch() {
  unsigned long sTime, dTime;
  bool initDone;
  
  displayStopwatchTime(0);
  initDone = false;
  
  // Wait until the rotary button is pressed and on release to start the stopwatch
  while (digitalRead(ROT_BUTTON_PIN));
  while (!digitalRead(ROT_BUTTON_PIN));         // WAIT until release
    

  while (digitalRead(ROT_BUTTON_PIN)) {         // WAIT until next press to stop the stopwatch
    if (!initDone) sTime = millis();
    initDone = true;

    // Read current time and define the delta
    dTime = millis() - sTime;

    displayStopwatchTime(dTime);
    delay(57);
  }   
  while (!digitalRead(ROT_BUTTON_PIN));         // WAIT until release

  // Keep the current stopwatch time visible until the next press, which exits the stopwatch function
  while (digitalRead(ROT_BUTTON_PIN));
  while (!digitalRead(ROT_BUTTON_PIN));         // WAIT until release
}

//--------------------------------------------------------------------------------------------------------------------
// Walk through the menu and select an item, returnvalue identifies the selection
// -------------------------------------------------------------------------------------------------------------------
char run_menu(char menuTxt[][10], char nrOfItems, bool normal) {
  char idx = 0;
  bool valueChanged;
  int R1Value, R1ValueOld;
  
  mx.clear();
  mx.setColumn(31, (normal)?0x3E:0x7C);
  mx.setColumn(30, (normal)?0x1C:0x38);
  mx.setColumn(29, (normal)?0x08:0x10);
  if (normal) {
    print_string(menuTxt[idx], 27);
  } else {
    print_string3(menuTxt[idx], 27, 2);
  }
  
  R1.Reset();
  R1ValueOld = 0;
  valueChanged = false;

  // As long as the button is not pressed, walk through the menu items, a button press selects the item
  while (digitalRead(ROT_BUTTON_PIN)) { 
    R1Value = R1.Read();
    if (R1.Changed()) {
      if (abs(R1Value - R1ValueOld) > 4) {
        if (R1Value > R1ValueOld) {
          if (idx < (nrOfItems-1)) idx++;
          valueChanged = true;
        } else {
          if (idx > 0) idx--;
          valueChanged = true;
        }
        R1ValueOld = R1Value; 

        mx.clear();
        mx.setColumn(31, (normal)?0x3E:0x7C);
        mx.setColumn(30, (normal)?0x1C:0x38);
        mx.setColumn(29, (normal)?0x08:0x10);
        if (normal) {
          print_string(menuTxt[idx], 27);
        } else {
          print_string3(menuTxt[idx], 27, 2);
        }
      }
    } 
  }
  while (!digitalRead(ROT_BUTTON_PIN));  // WAIT until release
  
  mx.clear();
  
  return idx;
}

//--------------------------------------------------------------------------------------------------------------------
// Execute the menu: Alarm items
// -------------------------------------------------------------------------------------------------------------------
void menu_alarm() {
  char menuItems[6][10] = {"On", "Off", "Time", "Tune", "Count", "Exit"};
  char select;

  select = run_menu(menuItems, 6, NORMALFONT);
  
  if (select == 0) set_alarm_onoff(1);  // Turn alarm on
  if (select == 1) set_alarm_onoff(0);  // Turn alarm off
  if (select == 2) set_alarm_time();    // Set the alarm time HH:MM:SS
  if (select == 3) set_sound(ALARM1);   // Set the alarm sound
  if (select == 4) set_alarm_count();   // Set the amount of sound repeats
  if (select == 5) ;                    // Exit
}

//--------------------------------------------------------------------------------------------------------------------
// Execute the menu: Alarm items
// -------------------------------------------------------------------------------------------------------------------
void menu_dark() {
  char menuItems[3][10] = {"On", "Off", "Exit"};
  char select;

  select = run_menu(menuItems, 3, NORMALFONT);
  
  if (select == 0) set_dark_onoff(1);  // Turn dark mode on
  if (select == 1) set_dark_onoff(0);  // Turn dark mode off
}

//--------------------------------------------------------------------------------------------------------------------
// Execute the menu: Settings items
// -------------------------------------------------------------------------------------------------------------------
void menu_settings() {
  char menuItems[6][10] = {"Tune", "Time", "Date", "Dark", "LCD", "Exit"};
  char select;

  select = run_menu(menuItems, 6, NORMALFONT);
  
  if (select == 0) set_sound(TIMER);    // Set the timer sound
  if (select == 1) set_time();          // Set the current time HH:MM:SS
  if (select == 2) set_date();          // Set the current date YYYY:MM:DD
  if (select == 3) menu_dark();         // Set or reset the dark display mode
  if (select == 4) set_lcd_intensity(); // Set the LCD intentsity
  if (select == 5) ;                    // Exit
}

//--------------------------------------------------------------------------------------------------------------------
// Execute the menu: Toplevel menu items
// -------------------------------------------------------------------------------------------------------------------
void process_menu() {
  char menuItems[4][10] = {"ALARM", "STOPWCH", "SETTING", "EXIT"};
  char select;

  select = run_menu(menuItems, 4, SMALLFONT);
  
  if (select == 0) menu_alarm();      // Go to the alarm sub menu
  if (select == 1) stopwatch();       // Run the stopwatch function
  if (select == 2) menu_settings();   // Go to the settings sub menu
  if (select == 3) ;                  // Exit

  mx.clear();
  m0p = 99;
}

//--------------------------------------------------------------------------------------------------------------------
// SETUP -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------
void setup() {
  byte alarmTime[4];                        // One byte for the hours, the minutes and the seconds
  unsigned char count;                      // Repetition count of the alarm
  uint32_t currentSecondsTime;              // Current time in seconds
  uint32_t storedSecondsTime;               // Stored time in seconds
  DateTime lastCorrectionTime, currentTime; // Used for checks if the clock needs to be corrected
  TimeSpan deltaTime;
  int totalCorrection;
  char buf[30];

  // Initialize the LCD
  mx.begin();
  mx.clear();
  mx.control(0, mx.INTENSITY, 1);
  mx.control(1, mx.INTENSITY, 1);
  mx.control(2, mx.INTENSITY, 1);
  mx.control(3, mx.INTENSITY, 1);
  
  // Initialize the serial interface over USB
  Serial.begin(57600);

  // Initialize the RTC
  RTCTime.init();

  // Initialize IO pin directions and pullups  
  pinMode(ROT_BUTTON_PIN, INPUT_PULLUP);
  pinMode(ROT_ENC1_PIN,   INPUT_PULLUP);
  pinMode(ROT_ENC2_PIN,   INPUT_PULLUP);
  pinMode(INTERRUPT_1HZ,  INPUT);
  pinMode(SOUND_OUTPUT,   OUTPUT);
  pinMode(SOUND_ON,       OUTPUT);
  digitalWrite(SOUND_OUTPUT, 0);
  digitalWrite(SOUND_ON, 0);
  
  // Initialize variables
  interrupt_1Hz_Flag = false;           // This boolean flag is set per 1Hz interrupt
  m0p = 99;                             // When set outside the range 0..59 it is guaranteed that HH:MM is also updated on the display
  time_mode = TIME_MODE_SMALL_HHMMSS;   // Selection :=={TIME_MODE_LARGE, TIME_MODE_SMALL_HHMM, TIME_MODE_SMALL_HHMMSS}
  timerSetTime1.h = 0;
  timerSetTime1.m = 2;
  timerSetTime1.s = 30;
  delay(500);
  
  // Initialize alarm
  alarm1_on = eeprom.read(ALARM1_ON_ADDRESS);
  if (alarm1_on != 1) alarm1_on = 0;    // Unitialized eeprom returns 0xff, only value 0x01 is 'on'
  if (alarm1_on == 1) {                 // When the alarm is turned on, fetch the alarm time
    // Read the alarm time and check if it's valid (valid = first byte equals 'ALARM_VALID_CODE')
    eeprom.read(ALARM1_ADDRESS, alarmTime, 4);
   
    if (alarmTime[0] == ALARM_VALID_CODE) {
      alarmSetTime1.h = alarmTime[1];
      alarmSetTime1.m = alarmTime[2];
      alarmSetTime1.s = alarmTime[3];
    } else {
      alarmSetTime1.h = 0;
      alarmSetTime1.m = 0;
      alarmSetTime1.s = 0;
    }    
  }

  // Initialize the alarm count, the very first time the default value is read which is 0xFF, this is changed to 5
  count = eeprom.read(ALARM1_REPEAT_ADDRESS);
  if (count > 250) eeprom.write(ALARM1_REPEAT_ADDRESS, 5);

  // Set correction factor (inaccuracy of RTC module in seconds per day)
  eeprom.write(CORRECTION_ADDRESS, 1);
  correction_done = false;
  correction_active = false;

  //read_display_eeprom_data();

  // Check if initial correction is needed, do this only when the correction is more than 1 day
  correction = eeprom.read(CORRECTION_ADDRESS);
  lastCorrectionTime = get_last_correction_time();
  currentTime = RTCTime.readCurrentTime();
  deltaTime = currentTime - lastCorrectionTime;
 
  if (deltaTime.days() > 0) {
    // Total correction is the correction per day multiplied by the amount of days the device was turned off
    totalCorrection = -correction * deltaTime.days();
    RTCTime.correctTime(totalCorrection);
    
    // Set the last known correction time to this time
    set_last_correction_time();
  }

  // Initialize Interrupt Service Routine
  attachInterrupt(digitalPinToInterrupt(INTERRUPT_1HZ), Int_1Hz_Routine, RISING);

  dark_on_countdown = 10;
  dark_on_and_show = false;
  
  randomSeed(millis());
  Serial.println("\nInit done");

  // For debugging purposes
  //read_display_eeprom_data();

  // Essential to wait 100ms to make sure the anti-glitch RC on the rotary push button becomes logical '1'
  delay(100);

  m0p = 99;
  mx.clear();
  print_time(time_mode);
}

//--------------------------------------------------------------------------------------------------------------------
// LOOP --------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------------
void loop() {
  bool button_pressed_twice;  
  char buf[30];
  
  // When the 1Hz flag is set, update the time and/or do other 1Hz related tasks
  if (interrupt_1Hz_Flag) {
    if (dark_on_and_show) {
      dark_on_countdown--;
      m0p = 99;
      if (dark_on_countdown <= 0) {
        dark_on_countdown = 10;
        dark_on_and_show = false;
        mx.clear();
      }
    }
    
    print_time(time_mode);
    interrupt_1Hz_Flag = false;
  }  

  // Check if rotary button is pressed
  if (!digitalRead(ROT_BUTTON_PIN)) {
    if ((dark_on == 1) && (!dark_on_and_show)) {
      // Dark mode: Button press turns on display
      dark_on_and_show = true;
      while (!digitalRead(ROT_BUTTON_PIN));         // WAIT until release
      
    } else {
      // Normal mode
      unsigned long time_count;
      button_pressed_twice = false;
      while (!digitalRead(ROT_BUTTON_PIN));         // WAIT until release
  
      // Check if the button is pressed again, within the next second
      time_count = millis();
      while ((millis() - time_count < 800) && (!button_pressed_twice)) {
        if (!digitalRead(ROT_BUTTON_PIN)) {
          button_pressed_twice = true;
        }
      }
      while (!digitalRead(ROT_BUTTON_PIN));         // WAIT until release
  
      // Main function is the kitchen timer, which is enabled by a single button press. Double press enters the menu.
      if (!button_pressed_twice) {
        process_timer();
      } else {
        process_menu();
      }
    }
  }
}
