//-----------------------------------------------------------------------------------------------
// Class to have standard calls whihc are mapped on a specific RTC, in this case the DS1307
// ----------------------------------------------------------------------------------------------
#include "Time.h"
#include "RTClib.h"

RTC_DS1307 rtc;

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Default constructor
//-----------------------------------------------------------------------------------------------------------------------------------------------
Time_Interface::Time_Interface() { 
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Initialize the RTC and set the SQ output to 1 Hz to use this signal as an interrupt trigger
//-----------------------------------------------------------------------------------------------------------------------------------------------
void Time_Interface::init() {
  // Start the timer and check availability
  if (!rtc.begin()) {
    Serial.println("RTC not found");
  }

  // Set the square wave output pin to 1Hz
  rtc.writeSqwPinMode(DS1307_SquareWave1HZ);

  // Due to whatever might happen the CH (Clock Hold) bit could be activated, resulting in a halted clock
  // Setting the SQW output pin to 1Hz is successful, but no signal is generated due to the Clock Halt
  // The RTClib 'should have' prevented this case by checking the CH bit and activate the clock when turned off
  // The lib does not provide means to activate the clock other than writing the seconds register which holds the CH bit
  // The next piece of code reads the CH bit, and when the clock is not running, we read the last known time and 
  // program it back into the clock, which activates the clock.
  char x = rtc.isrunning();
  if (x==0) {
    DateTime now = rtc.now();
    int sec = now.second();
    rtc.adjust(DateTime(now.year(), now.month(), now.day(), now.hour(), now.minute(), now.second()));
  }
  
  // Set date and time to the complile timestamp of this application
  //rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Read the current time
//-----------------------------------------------------------------------------------------------------------------------------------------------
DateTime Time_Interface::readCurrentTime() {
  return rtc.now();
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Correct the current time with the given amount of seconds, negative value is set clock backwards, positive value is set clock forwards
//-----------------------------------------------------------------------------------------------------------------------------------------------
void Time_Interface::correctTime(int secs) {
  char buf[60];

  // Read the current time
  DateTime now = rtc.now();
  sprintf(buf, "Old Time = %04d-%02d-%02d %02d:%02d:%02d", now.year(), now.month(), now.day(), now.hour(), now.minute(), now.second());
  Serial.println(buf);

  // Correct the read current time and write it back
  now = now + secs;
  rtc.adjust(now);
  
  // check
  now = rtc.now();
  sprintf(buf, "New Time = %04d-%02d-%02d %02d:%02d:%02d", now.year(), now.month(), now.day(), now.hour(), now.minute(), now.second());
  Serial.println(buf);
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Read time in passed seconds since 1 Jan 2000
//-----------------------------------------------------------------------------------------------------------------------------------------------
uint32_t Time_Interface::readSecondsTime() {
  uint32_t stime;
  DateTime now = rtc.now();
  stime = now.secondstime();
  return stime;
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Read time: Hour, Minutes, Seconds
//-----------------------------------------------------------------------------------------------------------------------------------------------
void Time_Interface::readHMS(int *h, int *m, int *s) {
  DateTime now = rtc.now();
  *h = now.hour();
  *m = now.minute();
  *s = now.second();
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Write time: Hour, Minutes, Seconds
//-----------------------------------------------------------------------------------------------------------------------------------------------
void Time_Interface::setHMS(int h, int m, int s) {
  DateTime now = rtc.now();
  rtc.adjust(DateTime(now.year(), now.month(), now.day(), h, m, s));
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Read date: Year, Month, Day
//-----------------------------------------------------------------------------------------------------------------------------------------------
void Time_Interface::readYMD(int *y, int *m, int *d) {
  DateTime now = rtc.now();
  *y = now.year();
  *m = now.month();
  *d = now.day();
}

//-----------------------------------------------------------------------------------------------------------------------------------------------
// Write date: Year, Month, Day
//-----------------------------------------------------------------------------------------------------------------------------------------------
void Time_Interface::setYMD(int y, int m, int d) {
  DateTime now = rtc.now();
  rtc.adjust(DateTime(y, m, d, now.hour(), now.minute(), now.second()));
}
